#include <linux/fs.h>
#include <linux/init.h>
#include <linux/miscdevice.h>
#include <linux/module.h>
#include <asm/uaccess.h>

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("\"Hello, world!\" minimal module");

static ssize_t hello_read(struct file * file, char * buf, 
			  size_t count, loff_t *ppos) {
	char *hello_str = "Hello, world!\n";
	int len = strlen(hello_str);
	/*
	 * We only support reading the whole string at once.
	 */
	if (count < len)
		return -EINVAL;
	/*
	 * If file position is non-zero, then assume the string has
	 * been read and indicate there is no more data to be read.
	 */
	if (*ppos != 0)
		return 0;
	if (copy_to_user(buf, hello_str, len))
		return -EINVAL;
	*ppos = len;
	return len;
}

static const struct file_operations hello_fops = {
	.owner		= THIS_MODULE,
	.read		= hello_read,
};

static struct miscdevice hello_dev = {
	/*
	 * kernel to just pick minor number.
	 */
	MISC_DYNAMIC_MINOR,
	/*
	 * Name ourselves /dev/hello.
	 */
	"hello",
	&hello_fops
};

static int __init
hello_init(void) {
	int ret;

	/*
	 * Create the "hello" device in the /sys/class/misc directory.
	 * Udev will automatically create the /dev/hello device using
	 * the default rules.
	 */
	ret = misc_register(&hello_dev);
	if (ret)
		printk(KERN_ERR
		       "Unable to register hello misc device\n");
	return ret;
}
module_init(hello_init);

static void __exit
hello_exit(void) {
	misc_deregister(&hello_dev);
}
module_exit(hello_exit);